package org.apache.bcel.verifier.statics;

import org.apache.bcel.classfile.Unknown;
import java.util.Locale;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.SourceFile;
import org.apache.bcel.classfile.InnerClasses;
import org.apache.bcel.classfile.InnerClass;
import org.apache.bcel.classfile.Synthetic;
import org.apache.bcel.classfile.ExceptionTable;
import org.apache.bcel.classfile.Deprecated;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.ConstantValue;
import org.apache.bcel.classfile.ConstantMethodref;
import org.apache.bcel.classfile.ConstantInterfaceMethodref;
import org.apache.bcel.classfile.ConstantFieldref;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.verifier.exc.LocalVariableInfoInconsistentException;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantDouble;
import org.apache.bcel.classfile.ConstantLong;
import org.apache.bcel.classfile.ConstantFloat;
import org.apache.bcel.classfile.ConstantInteger;
import org.apache.bcel.classfile.ConstantString;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.classfile.ConstantPool;
import org.apache.bcel.classfile.EmptyVisitor;
import org.apache.bcel.classfile.Method;
import java.util.Map;
import java.util.HashMap;
import org.apache.bcel.classfile.Visitor;
import org.apache.bcel.classfile.DescendingVisitor;
import java.util.Set;
import org.apache.bcel.verifier.VerifierFactory;
import org.apache.bcel.generic.Type;
import java.util.HashSet;
import org.apache.bcel.verifier.exc.ClassConstraintException;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.Repository;
import java.util.Objects;
import org.apache.bcel.classfile.Node;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.Constants;
import org.apache.bcel.verifier.PassVerifier;

public final class Pass2Verifier extends PassVerifier implements Constants
{
    private LocalVariablesInfo[] localVariablesInfos;
    private final Verifier myOwner;
    
    private static String tostring(final Node n) {
        return new StringRepresentation(n).toString();
    }
    
    private static boolean validClassMethodName(final String name) {
        return validMethodName(name, false);
    }
    
    private static boolean validClassName(final String name) {
        Objects.requireNonNull(name, "name");
        return true;
    }
    
    private static boolean validFieldName(final String name) {
        return validJavaIdentifier(name);
    }
    
    private static boolean validInterfaceMethodName(final String name) {
        return !name.startsWith("<") && validJavaLangMethodName(name);
    }
    
    private static boolean validJavaIdentifier(final String name) {
        if (name.isEmpty() || !Character.isJavaIdentifierStart(name.charAt(0))) {
            return false;
        }
        for (int i = 1; i < name.length(); ++i) {
            if (!Character.isJavaIdentifierPart(name.charAt(i))) {
                return false;
            }
        }
        return true;
    }
    
    private static boolean validJavaLangMethodName(final String name) {
        return validJavaIdentifier(name);
    }
    
    private static boolean validMethodName(final String name, final boolean allowStaticInit) {
        if (validJavaLangMethodName(name)) {
            return true;
        }
        if (allowStaticInit) {
            return name.equals("<init>") || name.equals("<clinit>");
        }
        return name.equals("<init>");
    }
    
    public Pass2Verifier(final Verifier owner) {
        this.myOwner = owner;
    }
    
    private void constantPoolEntriesSatisfyStaticConstraints() {
        try {
            final JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
            new CPESSC_Visitor(jc);
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    @Override
    public VerificationResult do_verify() {
        try {
            final VerificationResult vr1 = this.myOwner.doPass1();
            if (vr1.equals(VerificationResult.VR_OK)) {
                this.localVariablesInfos = new LocalVariablesInfo[Repository.lookupClass(this.myOwner.getClassName()).getMethods().length];
                VerificationResult vr2 = VerificationResult.VR_OK;
                try {
                    this.constantPoolEntriesSatisfyStaticConstraints();
                    this.fieldAndMethodRefsAreValid();
                    this.everyClassHasAnAccessibleSuperclass();
                    this.finalMethodsAreNotOverridden();
                }
                catch (final ClassConstraintException cce) {
                    vr2 = new VerificationResult(2, cce.getMessage());
                }
                return vr2;
            }
            return VerificationResult.VR_NOTYET;
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    private void everyClassHasAnAccessibleSuperclass() {
        try {
            final Set<String> hs = new HashSet<String>();
            JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
            int supidx = -1;
            while (supidx != 0) {
                supidx = jc.getSuperclassNameIndex();
                if (supidx == 0) {
                    if (jc != Repository.lookupClass(Type.OBJECT.getClassName())) {
                        throw new ClassConstraintException("Superclass of '" + jc.getClassName() + "' missing but not " + Type.OBJECT.getClassName() + " itself!");
                    }
                    continue;
                }
                else {
                    final String supername = jc.getSuperclassName();
                    if (!hs.add(supername)) {
                        throw new ClassConstraintException("Circular superclass hierarchy detected.");
                    }
                    final Verifier v = VerifierFactory.getVerifier(supername);
                    final VerificationResult vr = v.doPass1();
                    if (vr != VerificationResult.VR_OK) {
                        throw new ClassConstraintException("Could not load in ancestor class '" + supername + "'.");
                    }
                    jc = Repository.lookupClass(supername);
                    if (jc.isFinal()) {
                        throw new ClassConstraintException("Ancestor class '" + supername + "' has the FINAL access modifier and must therefore not be subclassed.");
                    }
                    continue;
                }
            }
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    private void fieldAndMethodRefsAreValid() {
        try {
            final JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
            final DescendingVisitor v = new DescendingVisitor(jc, new FAMRAV_Visitor(jc));
            v.visit();
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    private void finalMethodsAreNotOverridden() {
        try {
            final Map<String, String> map = new HashMap<String, String>();
            JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
            int supidx = -1;
            while (supidx != 0) {
                supidx = jc.getSuperclassNameIndex();
                final Method[] methods2;
                final Method[] methods = methods2 = jc.getMethods();
                for (final Method method : methods2) {
                    final String nameAndSig = method.getName() + method.getSignature();
                    if (map.containsKey(nameAndSig)) {
                        if (method.isFinal()) {
                            if (!method.isPrivate()) {
                                throw new ClassConstraintException("Method '" + nameAndSig + "' in class '" + map.get(nameAndSig) + "' overrides the final (not-overridable) definition in class '" + jc.getClassName() + "'.");
                            }
                            this.addMessage("Method '" + nameAndSig + "' in class '" + map.get(nameAndSig) + "' overrides the final (not-overridable) definition in class '" + jc.getClassName() + "'. This is okay, as the original definition was private; however this constraint leverage was introduced by JLS 8.4.6 (not vmspec2) and the behavior of the Sun verifiers.");
                        }
                        else if (!method.isStatic()) {
                            map.put(nameAndSig, jc.getClassName());
                        }
                    }
                    else if (!method.isStatic()) {
                        map.put(nameAndSig, jc.getClassName());
                    }
                }
                jc = Repository.lookupClass(jc.getSuperclassName());
            }
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    public LocalVariablesInfo getLocalVariablesInfo(final int methodNr) {
        if (this.verify() != VerificationResult.VR_OK) {
            return null;
        }
        if (methodNr < 0 || methodNr >= this.localVariablesInfos.length) {
            throw new AssertionViolatedException("Method number out of range.");
        }
        return this.localVariablesInfos[methodNr];
    }
    
    private final class CPESSC_Visitor extends EmptyVisitor
    {
        private final Class<?> CONST_Class;
        private final Class<?> CONST_String;
        private final Class<?> CONST_Integer;
        private final Class<?> CONST_Float;
        private final Class<?> CONST_Long;
        private final Class<?> CONST_Double;
        private final Class<?> CONST_NameAndType;
        private final Class<?> CONST_Utf8;
        private final JavaClass jc;
        private final ConstantPool cp;
        private final int cplen;
        private final DescendingVisitor carrier;
        private final Set<String> fieldNames;
        private final Set<String> fieldNamesAndDesc;
        private final Set<String> methodNamesAndDesc;
        
        private CPESSC_Visitor(final JavaClass jc) {
            this.fieldNames = new HashSet<String>();
            this.fieldNamesAndDesc = new HashSet<String>();
            this.methodNamesAndDesc = new HashSet<String>();
            this.jc = jc;
            this.cp = jc.getConstantPool();
            this.cplen = this.cp.getLength();
            this.CONST_Class = ConstantClass.class;
            this.CONST_String = ConstantString.class;
            this.CONST_Integer = ConstantInteger.class;
            this.CONST_Float = ConstantFloat.class;
            this.CONST_Long = ConstantLong.class;
            this.CONST_Double = ConstantDouble.class;
            this.CONST_NameAndType = ConstantNameAndType.class;
            this.CONST_Utf8 = ConstantUtf8.class;
            (this.carrier = new DescendingVisitor(jc, this)).visit();
        }
        
        private void checkIndex(final Node referrer, final int index, final Class<?> shouldbe) {
            if (index < 0 || index >= this.cplen) {
                throw new ClassConstraintException("Invalid index '" + index + "' used by '" + tostring(referrer) + "'.");
            }
            final Constant c = this.cp.getConstant(index);
            if (!shouldbe.isInstance(c)) {
                throw new ClassConstraintException("Illegal constant '" + tostring(c) + "' at index '" + index + "'. '" + tostring(referrer) + "' expects a '" + shouldbe + "'.");
            }
        }
        
        @Override
        public void visitCode(final Code obj) {
            try {
                this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
                final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
                if (!name.equals("Code")) {
                    throw new ClassConstraintException("The Code attribute '" + tostring(obj) + "' is not correctly named 'Code' but '" + name + "'.");
                }
                if (!(this.carrier.predecessor() instanceof Method)) {
                    Pass2Verifier.this.addMessage("Code attribute '" + tostring(obj) + "' is not declared in a method_info structure but in '" + this.carrier.predecessor() + "'. Ignored.");
                    return;
                }
                final Method m = (Method)this.carrier.predecessor();
                if (obj.getCode().length == 0) {
                    throw new ClassConstraintException("Code array of Code attribute '" + tostring(obj) + "' (method '" + m + "') must not be empty.");
                }
                final CodeException[] exceptionTable;
                final CodeException[] excTable = exceptionTable = obj.getExceptionTable();
                for (final CodeException element : exceptionTable) {
                    final int excIndex = element.getCatchType();
                    if (excIndex != 0) {
                        this.checkIndex(obj, excIndex, this.CONST_Class);
                        final ConstantClass cc = (ConstantClass)this.cp.getConstant(excIndex);
                        this.checkIndex(cc, cc.getNameIndex(), this.CONST_Utf8);
                        final String cname = Utility.pathToPackage(((ConstantUtf8)this.cp.getConstant(cc.getNameIndex())).getBytes());
                        Verifier v = VerifierFactory.getVerifier(cname);
                        VerificationResult vr = v.doPass1();
                        if (vr != VerificationResult.VR_OK) {
                            throw new ClassConstraintException("Code attribute '" + tostring(obj) + "' (method '" + m + "') has an exception_table entry '" + tostring(element) + "' that references '" + cname + "' as an Exception but it does not pass verification pass 1: " + vr);
                        }
                        JavaClass e = Repository.lookupClass(cname);
                        final JavaClass t = Repository.lookupClass(Type.THROWABLE.getClassName());
                        for (JavaClass o = Repository.lookupClass(Type.OBJECT.getClassName()); e != o && e != t; e = Repository.lookupClass(e.getSuperclassName())) {
                            v = VerifierFactory.getVerifier(e.getSuperclassName());
                            vr = v.doPass1();
                            if (vr != VerificationResult.VR_OK) {
                                throw new ClassConstraintException("Code attribute '" + tostring(obj) + "' (method '" + m + "') has an exception_table entry '" + tostring(element) + "' that references '" + cname + "' as an Exception but '" + e.getSuperclassName() + "' in the ancestor hierachy does not pass verification pass 1: " + vr);
                            }
                        }
                        if (e != t) {
                            throw new ClassConstraintException("Code attribute '" + tostring(obj) + "' (method '" + m + "') has an exception_table entry '" + tostring(element) + "' that references '" + cname + "' as an Exception but it is not a subclass of '" + t.getClassName() + "'.");
                        }
                    }
                }
                int methodNumber = -1;
                final Method[] ms = Repository.lookupClass(Pass2Verifier.this.myOwner.getClassName()).getMethods();
                for (int mn = 0; mn < ms.length; ++mn) {
                    if (m == ms[mn]) {
                        methodNumber = mn;
                        break;
                    }
                }
                if (methodNumber < 0) {
                    for (int mn = 0; mn < ms.length; ++mn) {
                        if (m.getName().equals(ms[mn].getName())) {
                            methodNumber = mn;
                            break;
                        }
                    }
                }
                if (methodNumber < 0) {
                    throw new AssertionViolatedException("Could not find a known BCEL Method object in the corresponding BCEL JavaClass object.");
                }
                Pass2Verifier.this.localVariablesInfos[methodNumber] = new LocalVariablesInfo(obj.getMaxLocals());
                int numOfLvtAttribs = 0;
                final Attribute[] attributes;
                final Attribute[] atts = attributes = obj.getAttributes();
                for (final Attribute att : attributes) {
                    if (!(att instanceof LineNumberTable) && !(att instanceof LocalVariableTable)) {
                        Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of Code attribute '" + tostring(obj) + "' (method '" + m + "') is unknown and will therefore be ignored.");
                    }
                    else {
                        Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of Code attribute '" + tostring(obj) + "' (method '" + m + "') will effectively be ignored and is only useful for debuggers and such.");
                    }
                    if (att instanceof LocalVariableTable) {
                        final LocalVariableTable lvt = (LocalVariableTable)att;
                        this.checkIndex(lvt, lvt.getNameIndex(), this.CONST_Utf8);
                        final String lvtname = ((ConstantUtf8)this.cp.getConstant(lvt.getNameIndex())).getBytes();
                        if (!lvtname.equals("LocalVariableTable")) {
                            throw new ClassConstraintException("The LocalVariableTable attribute '" + tostring(lvt) + "' is not correctly named 'LocalVariableTable' but '" + lvtname + "'.");
                        }
                        for (final LocalVariable localvariable : lvt.getLocalVariableTable()) {
                            this.checkIndex(lvt, localvariable.getNameIndex(), this.CONST_Utf8);
                            final String localname = ((ConstantUtf8)this.cp.getConstant(localvariable.getNameIndex())).getBytes();
                            if (!validJavaIdentifier(localname)) {
                                throw new ClassConstraintException("LocalVariableTable '" + tostring(lvt) + "' references a local variable by the name '" + localname + "' which is not a legal Java simple name.");
                            }
                            this.checkIndex(lvt, localvariable.getSignatureIndex(), this.CONST_Utf8);
                            final String localsig = ((ConstantUtf8)this.cp.getConstant(localvariable.getSignatureIndex())).getBytes();
                            Type t2;
                            try {
                                t2 = Type.getType(localsig);
                            }
                            catch (final ClassFormatException cfe) {
                                throw new ClassConstraintException("Illegal descriptor (==signature) '" + localsig + "' used by LocalVariable '" + tostring(localvariable) + "' referenced by '" + tostring(lvt) + "'.", cfe);
                            }
                            final int localindex = localvariable.getIndex();
                            if (((t2 == Type.LONG || t2 == Type.DOUBLE) ? (localindex + 1) : localindex) >= obj.getMaxLocals()) {
                                throw new ClassConstraintException("LocalVariableTable attribute '" + tostring(lvt) + "' references a LocalVariable '" + tostring(localvariable) + "' with an index that exceeds the surrounding Code attribute's max_locals value of '" + obj.getMaxLocals() + "'.");
                            }
                            try {
                                Pass2Verifier.this.localVariablesInfos[methodNumber].add(localindex, localname, localvariable.getStartPC(), localvariable.getLength(), t2);
                            }
                            catch (final LocalVariableInfoInconsistentException lviie) {
                                throw new ClassConstraintException("Conflicting information in LocalVariableTable '" + tostring(lvt) + "' found in Code attribute '" + tostring(obj) + "' (method '" + tostring(m) + "'). " + lviie.getMessage(), lviie);
                            }
                        }
                        ++numOfLvtAttribs;
                        if (!m.isStatic() && numOfLvtAttribs > obj.getMaxLocals()) {
                            throw new ClassConstraintException("Number of LocalVariableTable attributes of Code attribute '" + tostring(obj) + "' (method '" + tostring(m) + "') exceeds number of local variable slots '" + obj.getMaxLocals() + "' ('There may be at most one LocalVariableTable attribute per local variable in the Code attribute.').");
                        }
                    }
                }
            }
            catch (final ClassNotFoundException e2) {
                throw new AssertionViolatedException("Missing class: " + e2, e2);
            }
        }
        
        @Override
        public void visitCodeException(final CodeException obj) {
        }
        
        @Override
        public void visitConstantClass(final ConstantClass obj) {
            if (obj.getTag() != 7) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
        }
        
        @Override
        public void visitConstantDouble(final ConstantDouble obj) {
            if (obj.getTag() != 6) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
        }
        
        @Override
        public void visitConstantFieldref(final ConstantFieldref obj) {
            if (obj.getTag() != 9) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
            this.checkIndex(obj, obj.getClassIndex(), this.CONST_Class);
            this.checkIndex(obj, obj.getNameAndTypeIndex(), this.CONST_NameAndType);
        }
        
        @Override
        public void visitConstantFloat(final ConstantFloat obj) {
            if (obj.getTag() != 4) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
        }
        
        @Override
        public void visitConstantInteger(final ConstantInteger obj) {
            if (obj.getTag() != 3) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
        }
        
        @Override
        public void visitConstantInterfaceMethodref(final ConstantInterfaceMethodref obj) {
            if (obj.getTag() != 11) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
            this.checkIndex(obj, obj.getClassIndex(), this.CONST_Class);
            this.checkIndex(obj, obj.getNameAndTypeIndex(), this.CONST_NameAndType);
        }
        
        @Override
        public void visitConstantLong(final ConstantLong obj) {
            if (obj.getTag() != 5) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
        }
        
        @Override
        public void visitConstantMethodref(final ConstantMethodref obj) {
            if (obj.getTag() != 10) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
            this.checkIndex(obj, obj.getClassIndex(), this.CONST_Class);
            this.checkIndex(obj, obj.getNameAndTypeIndex(), this.CONST_NameAndType);
        }
        
        @Override
        public void visitConstantNameAndType(final ConstantNameAndType obj) {
            if (obj.getTag() != 12) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            this.checkIndex(obj, obj.getSignatureIndex(), this.CONST_Utf8);
        }
        
        @Override
        public void visitConstantPool(final ConstantPool obj) {
        }
        
        @Override
        public void visitConstantString(final ConstantString obj) {
            if (obj.getTag() != 8) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
            this.checkIndex(obj, obj.getStringIndex(), this.CONST_Utf8);
        }
        
        @Override
        public void visitConstantUtf8(final ConstantUtf8 obj) {
            if (obj.getTag() != 1) {
                throw new ClassConstraintException("Wrong constant tag in '" + tostring(obj) + "'.");
            }
        }
        
        @Override
        public void visitConstantValue(final ConstantValue obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
            if (!name.equals("ConstantValue")) {
                throw new ClassConstraintException("The ConstantValue attribute '" + tostring(obj) + "' is not correctly named 'ConstantValue' but '" + name + "'.");
            }
            final Object pred = this.carrier.predecessor();
            if (!(pred instanceof Field)) {
                return;
            }
            final Field f = (Field)pred;
            final Type fieldType = Type.getType(((ConstantUtf8)this.cp.getConstant(f.getSignatureIndex())).getBytes());
            final int index = obj.getConstantValueIndex();
            if (index < 0 || index >= this.cplen) {
                throw new ClassConstraintException("Invalid index '" + index + "' used by '" + tostring(obj) + "'.");
            }
            final Constant c = this.cp.getConstant(index);
            if ((this.CONST_Long.isInstance(c) && fieldType.equals(Type.LONG)) || (this.CONST_Float.isInstance(c) && fieldType.equals(Type.FLOAT))) {
                return;
            }
            if (this.CONST_Double.isInstance(c) && fieldType.equals(Type.DOUBLE)) {
                return;
            }
            if (this.CONST_Integer.isInstance(c) && (fieldType.equals(Type.INT) || fieldType.equals(Type.SHORT) || fieldType.equals(Type.CHAR) || fieldType.equals(Type.BYTE) || fieldType.equals(Type.BOOLEAN))) {
                return;
            }
            if (this.CONST_String.isInstance(c) && fieldType.equals(Type.STRING)) {
                return;
            }
            throw new ClassConstraintException("Illegal type of ConstantValue '" + obj + "' embedding Constant '" + c + "'. It is referenced by field '" + tostring(f) + "' expecting a different type: '" + fieldType + "'.");
        }
        
        @Override
        public void visitDeprecated(final Deprecated obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
            if (!name.equals("Deprecated")) {
                throw new ClassConstraintException("The Deprecated attribute '" + tostring(obj) + "' is not correctly named 'Deprecated' but '" + name + "'.");
            }
        }
        
        @Override
        public void visitExceptionTable(final ExceptionTable obj) {
            try {
                this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
                final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
                if (!name.equals("Exceptions")) {
                    throw new ClassConstraintException("The Exceptions attribute '" + tostring(obj) + "' is not correctly named 'Exceptions' but '" + name + "'.");
                }
                final int[] exceptionIndexTable;
                final int[] excIndices = exceptionIndexTable = obj.getExceptionIndexTable();
                for (final int excIndice : exceptionIndexTable) {
                    this.checkIndex(obj, excIndice, this.CONST_Class);
                    final ConstantClass cc = (ConstantClass)this.cp.getConstant(excIndice);
                    this.checkIndex(cc, cc.getNameIndex(), this.CONST_Utf8);
                    final String cname = Utility.pathToPackage(((ConstantUtf8)this.cp.getConstant(cc.getNameIndex())).getBytes());
                    Verifier v = VerifierFactory.getVerifier(cname);
                    VerificationResult vr = v.doPass1();
                    if (vr != VerificationResult.VR_OK) {
                        throw new ClassConstraintException("Exceptions attribute '" + tostring(obj) + "' references '" + cname + "' as an Exception but it does not pass verification pass 1: " + vr);
                    }
                    JavaClass e = Repository.lookupClass(cname);
                    final JavaClass t = Repository.lookupClass(Type.THROWABLE.getClassName());
                    for (JavaClass o = Repository.lookupClass(Type.OBJECT.getClassName()); e != o && e != t; e = Repository.lookupClass(e.getSuperclassName())) {
                        v = VerifierFactory.getVerifier(e.getSuperclassName());
                        vr = v.doPass1();
                        if (vr != VerificationResult.VR_OK) {
                            throw new ClassConstraintException("Exceptions attribute '" + tostring(obj) + "' references '" + cname + "' as an Exception but '" + e.getSuperclassName() + "' in the ancestor hierachy does not pass verification pass 1: " + vr);
                        }
                    }
                    if (e != t) {
                        throw new ClassConstraintException("Exceptions attribute '" + tostring(obj) + "' references '" + cname + "' as an Exception but it is not a subclass of '" + t.getClassName() + "'.");
                    }
                }
            }
            catch (final ClassNotFoundException e2) {
                throw new AssertionViolatedException("Missing class: " + e2, e2);
            }
        }
        
        @Override
        public void visitField(final Field obj) {
            if (this.jc.isClass()) {
                int maxone = 0;
                if (obj.isPrivate()) {
                    ++maxone;
                }
                if (obj.isProtected()) {
                    ++maxone;
                }
                if (obj.isPublic()) {
                    ++maxone;
                }
                if (maxone > 1) {
                    throw new ClassConstraintException("Field '" + tostring(obj) + "' must only have at most one of its ACC_PRIVATE, ACC_PROTECTED, ACC_PUBLIC modifiers set.");
                }
                if (obj.isFinal() && obj.isVolatile()) {
                    throw new ClassConstraintException("Field '" + tostring(obj) + "' must only have at most one of its ACC_FINAL, ACC_VOLATILE modifiers set.");
                }
            }
            else {
                if (!obj.isPublic()) {
                    throw new ClassConstraintException("Interface field '" + tostring(obj) + "' must have the ACC_PUBLIC modifier set but hasn't!");
                }
                if (!obj.isStatic()) {
                    throw new ClassConstraintException("Interface field '" + tostring(obj) + "' must have the ACC_STATIC modifier set but hasn't!");
                }
                if (!obj.isFinal()) {
                    throw new ClassConstraintException("Interface field '" + tostring(obj) + "' must have the ACC_FINAL modifier set but hasn't!");
                }
            }
            if ((obj.getAccessFlags() & 0xFFFFFF20) > 0) {
                Pass2Verifier.this.addMessage("Field '" + tostring(obj) + "' has access flag(s) other than ACC_PUBLIC, ACC_PRIVATE, ACC_PROTECTED, ACC_STATIC, ACC_FINAL, ACC_VOLATILE, ACC_TRANSIENT set (ignored).");
            }
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = obj.getName();
            if (!validFieldName(name)) {
                throw new ClassConstraintException("Field '" + tostring(obj) + "' has illegal name '" + obj.getName() + "'.");
            }
            this.checkIndex(obj, obj.getSignatureIndex(), this.CONST_Utf8);
            final String sig = ((ConstantUtf8)this.cp.getConstant(obj.getSignatureIndex())).getBytes();
            try {
                Type.getType(sig);
            }
            catch (final ClassFormatException cfe) {
                throw new ClassConstraintException("Illegal descriptor (==signature) '" + sig + "' used by '" + tostring(obj) + "'.", cfe);
            }
            final String nameanddesc = name + sig;
            if (this.fieldNamesAndDesc.contains(nameanddesc)) {
                throw new ClassConstraintException("No two fields (like '" + tostring(obj) + "') are allowed have same names and descriptors!");
            }
            if (this.fieldNames.contains(name)) {
                Pass2Verifier.this.addMessage("More than one field of name '" + name + "' detected (but with different type descriptors). This is very unusual.");
            }
            this.fieldNamesAndDesc.add(nameanddesc);
            this.fieldNames.add(name);
            final Attribute[] attributes;
            final Attribute[] atts = attributes = obj.getAttributes();
            for (final Attribute att : attributes) {
                if (!(att instanceof ConstantValue) && !(att instanceof Synthetic) && !(att instanceof Deprecated)) {
                    Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of Field '" + tostring(obj) + "' is unknown and will therefore be ignored.");
                }
                if (!(att instanceof ConstantValue)) {
                    Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of Field '" + tostring(obj) + "' is not a ConstantValue and is therefore only of use for debuggers and such.");
                }
            }
        }
        
        @Override
        public void visitInnerClass(final InnerClass obj) {
        }
        
        @Override
        public void visitInnerClasses(final InnerClasses innerClasses) {
            this.checkIndex(innerClasses, innerClasses.getNameIndex(), this.CONST_Utf8);
            final String name = ((ConstantUtf8)this.cp.getConstant(innerClasses.getNameIndex())).getBytes();
            if (!name.equals("InnerClasses")) {
                throw new ClassConstraintException("The InnerClasses attribute '" + tostring(innerClasses) + "' is not correctly named 'InnerClasses' but '" + name + "'.");
            }
            innerClasses.forEach(ic -> {
                this.checkIndex(innerClasses, ic.getInnerClassIndex(), this.CONST_Class);
                final int outerIdx = ic.getOuterClassIndex();
                if (outerIdx != 0) {
                    this.checkIndex(innerClasses, outerIdx, this.CONST_Class);
                }
                final int innernameIdx = ic.getInnerNameIndex();
                if (innernameIdx != 0) {
                    this.checkIndex(innerClasses, innernameIdx, this.CONST_Utf8);
                }
                final int acc = ic.getInnerAccessFlags();
                final int acc2 = acc & 0xFFFFF9E0;
                if (acc2 != 0) {
                    Pass2Verifier.this.addMessage("Unknown access flag for inner class '" + tostring(ic) + "' set (InnerClasses attribute '" + tostring(innerClasses) + "').");
                }
            });
        }
        
        @Override
        public void visitJavaClass(final JavaClass obj) {
            final Attribute[] atts = obj.getAttributes();
            boolean foundSourceFile = false;
            boolean foundInnerClasses = false;
            final boolean hasInnerClass = new InnerClassDetector(this.jc).innerClassReferenced();
            for (final Attribute att : atts) {
                if (!(att instanceof SourceFile) && !(att instanceof Deprecated) && !(att instanceof InnerClasses) && !(att instanceof Synthetic)) {
                    Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of the ClassFile structure '" + tostring(obj) + "' is unknown and will therefore be ignored.");
                }
                if (att instanceof SourceFile) {
                    if (foundSourceFile) {
                        throw new ClassConstraintException("A ClassFile structure (like '" + tostring(obj) + "') may have no more than one SourceFile attribute.");
                    }
                    foundSourceFile = true;
                }
                if (att instanceof InnerClasses) {
                    if (!foundInnerClasses) {
                        foundInnerClasses = true;
                    }
                    else if (hasInnerClass) {
                        throw new ClassConstraintException("A Classfile structure (like '" + tostring(obj) + "') must have exactly one InnerClasses attribute if at least one Inner Class is referenced (which is the case). More than one InnerClasses attribute was found.");
                    }
                    if (!hasInnerClass) {
                        Pass2Verifier.this.addMessage("No referenced Inner Class found, but InnerClasses attribute '" + tostring(att) + "' found. Strongly suggest removal of that attribute.");
                    }
                }
            }
            if (hasInnerClass && !foundInnerClasses) {
                Pass2Verifier.this.addMessage("A Classfile structure (like '" + tostring(obj) + "') must have exactly one InnerClasses attribute if at least one Inner Class is referenced (which is the case). No InnerClasses attribute was found.");
            }
        }
        
        @Override
        public void visitLineNumber(final LineNumber obj) {
        }
        
        @Override
        public void visitLineNumberTable(final LineNumberTable obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
            if (!name.equals("LineNumberTable")) {
                throw new ClassConstraintException("The LineNumberTable attribute '" + tostring(obj) + "' is not correctly named 'LineNumberTable' but '" + name + "'.");
            }
        }
        
        @Override
        public void visitLocalVariable(final LocalVariable obj) {
        }
        
        @Override
        public void visitLocalVariableTable(final LocalVariableTable obj) {
        }
        
        @Override
        public void visitMethod(final Method obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = obj.getName();
            if (!validMethodName(name, true)) {
                throw new ClassConstraintException("Method '" + tostring(obj) + "' has illegal name '" + name + "'.");
            }
            this.checkIndex(obj, obj.getSignatureIndex(), this.CONST_Utf8);
            final String sig = ((ConstantUtf8)this.cp.getConstant(obj.getSignatureIndex())).getBytes();
            Type t;
            Type[] ts;
            try {
                t = Type.getReturnType(sig);
                ts = Type.getArgumentTypes(sig);
            }
            catch (final ClassFormatException cfe) {
                throw new ClassConstraintException("Illegal descriptor (==signature) '" + sig + "' used by Method '" + tostring(obj) + "'.", cfe);
            }
            Type act = t;
            if (act instanceof ArrayType) {
                act = ((ArrayType)act).getBasicType();
            }
            if (act instanceof ObjectType) {
                final Verifier v = VerifierFactory.getVerifier(((ObjectType)act).getClassName());
                final VerificationResult vr = v.doPass1();
                if (vr != VerificationResult.VR_OK) {
                    throw new ClassConstraintException("Method '" + tostring(obj) + "' has a return type that does not pass verification pass 1: '" + vr + "'.");
                }
            }
            final Type[] array = ts;
            for (int length = array.length, i = 0; i < length; ++i) {
                final Type element = act = array[i];
                if (act instanceof ArrayType) {
                    act = ((ArrayType)act).getBasicType();
                }
                if (act instanceof ObjectType) {
                    final Verifier v2 = VerifierFactory.getVerifier(((ObjectType)act).getClassName());
                    final VerificationResult vr2 = v2.doPass1();
                    if (vr2 != VerificationResult.VR_OK) {
                        throw new ClassConstraintException("Method '" + tostring(obj) + "' has an argument type that does not pass verification pass 1: '" + vr2 + "'.");
                    }
                }
            }
            if (name.equals("<clinit>") && ts.length != 0) {
                throw new ClassConstraintException("Method '" + tostring(obj) + "' has illegal name '" + name + "'. Its name resembles the class or interface initialization method which it isn't because of its arguments (==descriptor).");
            }
            if (this.jc.isClass()) {
                int maxone = 0;
                if (obj.isPrivate()) {
                    ++maxone;
                }
                if (obj.isProtected()) {
                    ++maxone;
                }
                if (obj.isPublic()) {
                    ++maxone;
                }
                if (maxone > 1) {
                    throw new ClassConstraintException("Method '" + tostring(obj) + "' must only have at most one of its ACC_PRIVATE, ACC_PROTECTED, ACC_PUBLIC modifiers set.");
                }
                if (obj.isAbstract()) {
                    if (obj.isFinal()) {
                        throw new ClassConstraintException("Abstract method '" + tostring(obj) + "' must not have the ACC_FINAL modifier set.");
                    }
                    if (obj.isNative()) {
                        throw new ClassConstraintException("Abstract method '" + tostring(obj) + "' must not have the ACC_NATIVE modifier set.");
                    }
                    if (obj.isPrivate()) {
                        throw new ClassConstraintException("Abstract method '" + tostring(obj) + "' must not have the ACC_PRIVATE modifier set.");
                    }
                    if (obj.isStatic()) {
                        throw new ClassConstraintException("Abstract method '" + tostring(obj) + "' must not have the ACC_STATIC modifier set.");
                    }
                    if (obj.isStrictfp()) {
                        throw new ClassConstraintException("Abstract method '" + tostring(obj) + "' must not have the ACC_STRICT modifier set.");
                    }
                    if (obj.isSynchronized()) {
                        throw new ClassConstraintException("Abstract method '" + tostring(obj) + "' must not have the ACC_SYNCHRONIZED modifier set.");
                    }
                }
                if (name.equals("<init>") && (obj.isStatic() || obj.isFinal() || obj.isSynchronized() || obj.isNative() || obj.isAbstract())) {
                    throw new ClassConstraintException("Instance initialization method '" + tostring(obj) + "' must not have any of the ACC_STATIC, ACC_FINAL, ACC_SYNCHRONIZED, ACC_NATIVE, ACC_ABSTRACT modifiers set.");
                }
            }
            else if (!name.equals("<clinit>")) {
                if (this.jc.getMajor() >= 52) {
                    if (obj.isPublic() == obj.isPrivate()) {
                        throw new ClassConstraintException("Interface method '" + tostring(obj) + "' must have exactly one of its ACC_PUBLIC and ACC_PRIVATE modifiers set.");
                    }
                    if (obj.isProtected() || obj.isFinal() || obj.isSynchronized() || obj.isNative()) {
                        throw new ClassConstraintException("Interface method '" + tostring(obj) + "' must not have any of the ACC_PROTECTED, ACC_FINAL, ACC_SYNCHRONIZED, or ACC_NATIVE modifiers set.");
                    }
                }
                else {
                    if (!obj.isPublic()) {
                        throw new ClassConstraintException("Interface method '" + tostring(obj) + "' must have the ACC_PUBLIC modifier set but hasn't!");
                    }
                    if (!obj.isAbstract()) {
                        throw new ClassConstraintException("Interface method '" + tostring(obj) + "' must have the ACC_ABSTRACT modifier set but hasn't!");
                    }
                    if (obj.isPrivate() || obj.isProtected() || obj.isStatic() || obj.isFinal() || obj.isSynchronized() || obj.isNative() || obj.isStrictfp()) {
                        throw new ClassConstraintException("Interface method '" + tostring(obj) + "' must not have any of the ACC_PRIVATE, ACC_PROTECTED, ACC_STATIC, ACC_FINAL, ACC_SYNCHRONIZED, ACC_NATIVE, ACC_ABSTRACT, ACC_STRICT modifiers set.");
                    }
                }
            }
            if ((obj.getAccessFlags() & 0xFFFFF2C0) > 0) {
                Pass2Verifier.this.addMessage("Method '" + tostring(obj) + "' has access flag(s) other than ACC_PUBLIC, ACC_PRIVATE, ACC_PROTECTED, ACC_STATIC, ACC_FINAL, ACC_SYNCHRONIZED, ACC_NATIVE, ACC_ABSTRACT, ACC_STRICT set (ignored).");
            }
            final String nameanddesc = name + sig;
            if (this.methodNamesAndDesc.contains(nameanddesc)) {
                throw new ClassConstraintException("No two methods (like '" + tostring(obj) + "') are allowed have same names and desciptors!");
            }
            this.methodNamesAndDesc.add(nameanddesc);
            final Attribute[] atts = obj.getAttributes();
            int numCodeAtts = 0;
            for (final Attribute att : atts) {
                if (!(att instanceof Code) && !(att instanceof ExceptionTable) && !(att instanceof Synthetic) && !(att instanceof Deprecated)) {
                    Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of Method '" + tostring(obj) + "' is unknown and will therefore be ignored.");
                }
                if (!(att instanceof Code) && !(att instanceof ExceptionTable)) {
                    Pass2Verifier.this.addMessage("Attribute '" + tostring(att) + "' as an attribute of Method '" + tostring(obj) + "' is neither Code nor Exceptions and is therefore only of use for debuggers and such.");
                }
                if (att instanceof Code && (obj.isNative() || obj.isAbstract())) {
                    throw new ClassConstraintException("Native or abstract methods like '" + tostring(obj) + "' must not have a Code attribute like '" + tostring(att) + "'.");
                }
                if (att instanceof Code) {
                    ++numCodeAtts;
                }
            }
            if (!obj.isNative() && !obj.isAbstract() && numCodeAtts != 1) {
                throw new ClassConstraintException("Non-native, non-abstract methods like '" + tostring(obj) + "' must have exactly one Code attribute (found: " + numCodeAtts + ").");
            }
        }
        
        @Override
        public void visitSourceFile(final SourceFile obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
            if (!name.equals("SourceFile")) {
                throw new ClassConstraintException("The SourceFile attribute '" + tostring(obj) + "' is not correctly named 'SourceFile' but '" + name + "'.");
            }
            this.checkIndex(obj, obj.getSourceFileIndex(), this.CONST_Utf8);
            final String sourceFileName = ((ConstantUtf8)this.cp.getConstant(obj.getSourceFileIndex())).getBytes();
            final String sourceFileNameLc = sourceFileName.toLowerCase(Locale.ENGLISH);
            if (sourceFileName.indexOf(47) != -1 || sourceFileName.indexOf(92) != -1 || sourceFileName.indexOf(58) != -1 || sourceFileNameLc.lastIndexOf(".java") == -1) {
                Pass2Verifier.this.addMessage("SourceFile attribute '" + tostring(obj) + "' has a funny name: remember not to confuse certain parsers working on javap's output. Also, this name ('" + sourceFileName + "') is considered an unqualified (simple) file name only.");
            }
        }
        
        @Override
        public void visitSynthetic(final Synthetic obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            final String name = ((ConstantUtf8)this.cp.getConstant(obj.getNameIndex())).getBytes();
            if (!name.equals("Synthetic")) {
                throw new ClassConstraintException("The Synthetic attribute '" + tostring(obj) + "' is not correctly named 'Synthetic' but '" + name + "'.");
            }
        }
        
        @Override
        public void visitUnknown(final Unknown obj) {
            this.checkIndex(obj, obj.getNameIndex(), this.CONST_Utf8);
            Pass2Verifier.this.addMessage("Unknown attribute '" + tostring(obj) + "'. This attribute is not known in any context!");
        }
    }
    
    private final class FAMRAV_Visitor extends EmptyVisitor
    {
        private final ConstantPool cp;
        
        private FAMRAV_Visitor(final JavaClass jc) {
            this.cp = jc.getConstantPool();
        }
        
        @Override
        public void visitConstantFieldref(final ConstantFieldref obj) {
            if (obj.getTag() != 9) {
                throw new ClassConstraintException("ConstantFieldref '" + tostring(obj) + "' has wrong tag!");
            }
            final int nameAndTypeIndex = obj.getNameAndTypeIndex();
            final ConstantNameAndType cnat = (ConstantNameAndType)this.cp.getConstant(nameAndTypeIndex);
            final String name = ((ConstantUtf8)this.cp.getConstant(cnat.getNameIndex())).getBytes();
            if (!validFieldName(name)) {
                throw new ClassConstraintException("Invalid field name '" + name + "' referenced by '" + tostring(obj) + "'.");
            }
            final int classIndex = obj.getClassIndex();
            final ConstantClass cc = (ConstantClass)this.cp.getConstant(classIndex);
            final String className = ((ConstantUtf8)this.cp.getConstant(cc.getNameIndex())).getBytes();
            if (!validClassName(className)) {
                throw new ClassConstraintException("Illegal class name '" + className + "' used by '" + tostring(obj) + "'.");
            }
            final String sig = ((ConstantUtf8)this.cp.getConstant(cnat.getSignatureIndex())).getBytes();
            try {
                Type.getType(sig);
            }
            catch (final ClassFormatException cfe) {
                throw new ClassConstraintException("Illegal descriptor (==signature) '" + sig + "' used by '" + tostring(obj) + "'.", cfe);
            }
        }
        
        @Override
        public void visitConstantInterfaceMethodref(final ConstantInterfaceMethodref obj) {
            if (obj.getTag() != 11) {
                throw new ClassConstraintException("ConstantInterfaceMethodref '" + tostring(obj) + "' has wrong tag!");
            }
            final int nameAndTypeIndex = obj.getNameAndTypeIndex();
            final ConstantNameAndType cnat = (ConstantNameAndType)this.cp.getConstant(nameAndTypeIndex);
            final String name = ((ConstantUtf8)this.cp.getConstant(cnat.getNameIndex())).getBytes();
            if (!validInterfaceMethodName(name)) {
                throw new ClassConstraintException("Invalid (interface) method name '" + name + "' referenced by '" + tostring(obj) + "'.");
            }
            final int classIndex = obj.getClassIndex();
            final ConstantClass cc = (ConstantClass)this.cp.getConstant(classIndex);
            final String className = ((ConstantUtf8)this.cp.getConstant(cc.getNameIndex())).getBytes();
            if (!validClassName(className)) {
                throw new ClassConstraintException("Illegal class name '" + className + "' used by '" + tostring(obj) + "'.");
            }
            final String sig = ((ConstantUtf8)this.cp.getConstant(cnat.getSignatureIndex())).getBytes();
            try {
                final Type t = Type.getReturnType(sig);
                if (name.equals("<clinit>") && t != Type.VOID) {
                    Pass2Verifier.this.addMessage("Class or interface initialization method '<clinit>' usually has VOID return type instead of '" + t + "'. Note this is really not a requirement of The Java Virtual Machine Specification, Second Edition.");
                }
            }
            catch (final ClassFormatException cfe) {
                throw new ClassConstraintException("Illegal descriptor (==signature) '" + sig + "' used by '" + tostring(obj) + "'.", cfe);
            }
        }
        
        @Override
        public void visitConstantMethodref(final ConstantMethodref obj) {
            if (obj.getTag() != 10) {
                throw new ClassConstraintException("ConstantMethodref '" + tostring(obj) + "' has wrong tag!");
            }
            final int nameAndTypeIndex = obj.getNameAndTypeIndex();
            final ConstantNameAndType cnat = (ConstantNameAndType)this.cp.getConstant(nameAndTypeIndex);
            final String name = ((ConstantUtf8)this.cp.getConstant(cnat.getNameIndex())).getBytes();
            if (!validClassMethodName(name)) {
                throw new ClassConstraintException("Invalid (non-interface) method name '" + name + "' referenced by '" + tostring(obj) + "'.");
            }
            final int classIndex = obj.getClassIndex();
            final ConstantClass cc = (ConstantClass)this.cp.getConstant(classIndex);
            final String className = ((ConstantUtf8)this.cp.getConstant(cc.getNameIndex())).getBytes();
            if (!validClassName(className)) {
                throw new ClassConstraintException("Illegal class name '" + className + "' used by '" + tostring(obj) + "'.");
            }
            final String sig = ((ConstantUtf8)this.cp.getConstant(cnat.getSignatureIndex())).getBytes();
            try {
                final Type t = Type.getReturnType(sig);
                if (name.equals("<init>") && t != Type.VOID) {
                    throw new ClassConstraintException("Instance initialization method must have VOID return type.");
                }
            }
            catch (final ClassFormatException cfe) {
                throw new ClassConstraintException("Illegal descriptor (==signature) '" + sig + "' used by '" + tostring(obj) + "'.", cfe);
            }
        }
    }
    
    private static class InnerClassDetector extends EmptyVisitor
    {
        private boolean hasInnerClass;
        private final JavaClass jc;
        private final ConstantPool cp;
        
        public InnerClassDetector(final JavaClass javaClass) {
            this.jc = javaClass;
            this.cp = this.jc.getConstantPool();
            new DescendingVisitor(this.jc, this).visit();
        }
        
        public boolean innerClassReferenced() {
            return this.hasInnerClass;
        }
        
        @Override
        public void visitConstantClass(final ConstantClass obj) {
            final Constant c = this.cp.getConstant(obj.getNameIndex());
            if (c instanceof ConstantUtf8) {
                final String className = ((ConstantUtf8)c).getBytes();
                if (className.startsWith(this.jc.getClassName().replace('.', '/') + "$")) {
                    this.hasInnerClass = true;
                }
            }
        }
    }
}
